from .matcher import GeoMatcher
from pyarrow import fs
import pyarrow.parquet as pq
import numpy as np
from urllib import parse
from collections import Counter
from PIL import Image
import requests
import io


class Analysis(GeoMatcher):
    def get_map_data_by_address(
        self,
        address,
        apikey,
        similarity_threshold=0.9,
        top_result=True,
        regions=[],
        operator=None,
        region="",
        address_cleaning=False,
        string_metric="levenshtein",
    ):

        geocode_dict = self._get_geocode_by_address(
            address,
            similarity_threshold=similarity_threshold,
            top_result=top_result,
            regions=regions,
            operator=operator,
            region=region,
            address_cleaning=address_cleaning,
            string_metric=string_metric,
        )

        img = self._get_static_map_by_coord(
            geocode_dict["LATITUDE"], geocode_dict["LONGITUDE"], apikey
        )
        if not img:
            return None
        return self._map_analysis(img)

    def _get_static_map_by_coord(
        self, lat, lon, apikey, zoom=13, size=(400, 400), maptype="roadmap"
    ):
        style_dict = {
            "landscape.man_made": "#CC9999",
            "landscape.natural": "#66FF00",
            "poi.attraction": "#99CCFF",
            "poi.business": "#99FF99",
            "poi.medical": "#FF0000",
            "poi.park": "#009900",
            "poi.school": "#CC0033",
            "poi.sports_complex": "#993300",
            "road.arterial": "#669999",
            "road.highway": "#3399CC",
            "road.local": "#FFCCCC",
            "transit.line": "#66FFCC",
            "transit.station.airport": "#FFCC99",
            "transit.station.bus": "#FFFF99",
            "transit.station.rail": "#800080",
            "water": "#00FFFF",
        }

        url = "https://maps.googleapis.com/maps/api/staticmap"
        assert maptype in (
            "roadmap",
            "satellite",
            "hybrid",
            "terrain",
        ), "maptype must be one of 'roadmap', 'statelite', 'hybrid' and 'terrain'"

        params = {
            "center": f"{lat},{lon}",
            "zoom": zoom,
            "maptype": maptype,
            "size": f"{size[0]}x{size[1]}",
            "key": apikey,
        }

        url += f"?{parse.urlencode(params)}"
        for style, colour in style_dict.items():
            url += f"&style=feature:{style}|color:0x{colour.lstrip('#').lower()}"
        r = requests.get(url, params=params)
        if r.status_code != 200:
            print(f"Status Code {r.status_code}")
            return
        img = Image.open(io.BytesIO(r.content)).convert("RGB")
        return img

    def _get_geocode_by_address(
        self,
        address,
        similarity_threshold=0.9,
        top_result=True,
        regions=[],
        operator=None,
        region="",
        address_cleaning=False,
        string_metric="levenshtein",
    ):

        address_df = self.get_region_by_address(
            address,
            similarity_threshold=similarity_threshold,
            top_result=top_result,
            regions=regions,
            operator=operator,
            region=region,
            address_cleaning=address_cleaning,
            string_metric=string_metric,
        )

        if not address_df:
            return
        addr_list = address_df.get("FULL_ADDRESS")
        if not addr_list:
            return
        addr = addr_list[0]
        local = fs.LocalFileSystem()
        geocode_dict = {
            k: v[0]
            for k, v in pq.read_table(
                self._filename,
                columns=["FULL_ADDRESS", "LATITUDE", "LONGITUDE"],
                filesystem=local,
                filters=[("FULL_ADDRESS", "==", addr)],
            )
            .to_pydict()
            .items()
        }
        return geocode_dict

    def _map_analysis(self, img):
        col_dict = {
            (203, 152, 152): "landscape.man_made",
            (102, 254, 0): "landscape.natural",
            (153, 204, 255): "poi.attraction",
            (153, 255, 153): "poi.business",
            (255, 0, 0): "poi.medical",
            (0, 153, 0): "poi.park",
            (204, 0, 51): "poi.school",
            (153, 51, 0): "poi.sports_complex",
            (102, 152, 152): "road.arterial",
            (51, 153, 203): "road.highway",
            (249, 198, 198): "road.local",
            (102, 253, 202): "transit.line",
            (254, 203, 153): "transit.station.airport",
            (255, 255, 153): "transit.station.bus",
            (128, 0, 128): "transit.station.rail",
            (0, 254, 254): "water",
        }
        img_array = np.asarray(img).reshape(-1, 3)
        total = img_array.shape[0]
        colors = dict(Counter(map(tuple, img_array)))
        area_dict = {}
        for ary, count in list(colors.items()):
            for k, v in col_dict.items():
                if ary in colors:
                    if np.sqrt(np.mean((np.array(ary) - np.array(k)) ** 2)) < 7:
                        area_dict[v] = area_dict[v] + count if v in area_dict else count
                        colors.pop(ary)
        area_dict = {k: v / total for k, v in area_dict.items()}
        if not area_dict:
            return
        return area_dict



from .region import Region
from .hierarchy import GeoHierarchy

country = Region("Country", "National")

# administrative level
state = Region("State", "State", "STATE")
lga = Region("Local Government Area", "LGA", "LGA_NAME_2016")
suburb = Region("Suburb", "Suburb", "SSC_NAME_2016")

# statistical area
sa4 = Region("Statistical Area 4", "SA4", "SA4_NAME_2016")
sa3 = Region("Statistical Area 3", "SA3", "SA3_NAME_2016")
sa2 = Region("Statistical Area 2", "SA2", "SA2_NAME_2016")
sa1 = Region("Statistical Area 1", "SA1", "SA1_7DIGITCODE_2016")

mb = Region("Meshblock", "MB", "MB_CODE_2016")

AUS = GeoHierarchy(country, "Australia")

AUS.add_region(region=state, parent_region=country)

# administrative
AUS.add_region(region=lga, parent_region=state)
AUS.add_region(region=suburb, parent_region=lga)
AUS.add_region(region=mb, parent_region=suburb)

# statistical area
AUS.add_region(region=sa4, parent_region=state)
AUS.add_region(region=sa3, parent_region=sa4)
AUS.add_region(region=sa2, parent_region=sa3)
AUS.add_region(region=sa1, parent_region=sa2)
AUS.add_region(region=mb, parent_region=sa1)

AUS.add_type(lga, "Administrative")
AUS.add_type(sa4, "ASGS", "Australian Statistical Geography Standard")

AUS.coordinate_boundary = [-43.58301104, -9.23000371,
                           96.82159219, 167.99384663]



from .region import Region
from operator import le, ge


class GeoHierarchy:
    """
    The GeoHierarchy class represents the structure of a country's region/area;
    for instance, a state or a province is the sub-region of a country.
    """

    __slot__ = (
        "_type",
        "_type_root",
        "_country",
        "_country_name",
        "_coordinate_boundary",
    )

    def __init__(self, country, name, coordinate_boundary=None):
        # country is the root of the hierarchy
        self._type = {}
        self._type_root = {}
        self._country = self._Node(country)
        self._country_name = name
        self.coordinate_boundary = coordinate_boundary

    @property
    def type(self):
        """
        Return the dictionary that contains all the defined regional structures.
        Returns
        -------
        dictionary
            The defined regional structures
        Examples
        --------
        >>> country = Region("Country")
        >>> state = Region("State",col_name="STATE")
        >>> sa4 = Region("Statistica Area 4",short_name="SA4",col_name="SA4")
        >>> australia = new GeoHierarchy(country,"Australia")
        >>> australia.add_region(region=state, parent_region=country)
        >>> australia.add_region(region=sa4, parent_region=state)
        >>> australia.add_type(sa4,"ASGS","Australian Statistical Geography Standard")
        >>> australia.type
        {'ASGS': 'Australian Statistical Geography Standard'}
        """
        return self._type

    @property
    def name(self):
        """
        Return the name of the country
        Returns
        -------
        string
            The name of the country
        Examples
        --------
        The country's name can be set initially when calling the constructor.
        >>> australia = new GeoHierarchy(country,"Australia")
        >>> australia.name
        'Australia'
        """
        return self._country_name

    @property
    def coordinate_boundary(self):
        """
        Return the list of coordinates as the boundary of a country.
        The format is [minimum latitude, maximum latitude, minimum longitude,
        maximum longitude]
        Returns
        -------
        string
            The coordinate boundary of a country
        Examples
        --------
        >>> australia = new GeoHierarchy(country,"Australia")
        >>> australia.coordinate_boundary = [-43.58301104, -9.23000371,
                                             96.82159219, 167.99384663]
        >>> australia.coordinate_boundary
        [-43.58301104, -9.23000371, 96.82159219, 167.99384663]
        """
        return self._coordinate_boundary

    @coordinate_boundary.setter
    def coordinate_boundary(self, value):
        """
        Set or modify the coordinate boundaries of a country.
        Parameters
        ----------
        value:list
            The coordinate boundary of a country.
            The format of the input is : [minimum latitude, maximum latitude,
            minimum longitude, maximum longitude]
        Examples
        --------
        >>> australia = new GeoHierarchy(country,"Australia")
        >>> australia.coordinate_boundary = [-43.58301104, -9.23000371,
                                             96.82159219, 167.99384663]
        >>> australia.coordinate_boundary
        [-43.58301104, -9.23000371, 96.82159219, 167.99384663]
        """
        if value is not None:
            if isinstance(value, list) and len(value):
                if (value[0] < value[1]) and (value[2] < value[3]):
                    self._coordinate_boundary = value
                else:
                    if value[0] >= value[1]:
                        raise ValueError(
                            "The latitute range is invalid (lat_max >= lat_min)"
                        )
                    else:
                        raise ValueError(
                            "The longitude range is invalid (long_max >= long_min)"
                        )
            else:
                raise ValueError(
                    "The boundary must be a four element list "
                    "[lat_min, lat_max, long_min, long_max]"
                )

    def add_type(self, region, type_id, type_name=""):
        """
        Add a new geographical hierarchy type, for instace: statistical area,
        administrative level
        Parameters
        ----------
        region:Region
            The smallest root region for the hierarchy.
            The common upper-level region name, shared with other types of
            hierarchies, can't be assigned as a root.
            For instance, administrative level and statistical area use
            Country or State/Province as their upper regional level
        type_id:string
            The unique identifier for the structural type
        type_name:string
            The name of the regional structure type
        Examples
        --------
        >>> country = Region("Country")
        >>> state = Region("State",col_name="STATE")
        >>> sa4 = Region("Statistica Area 4",short_name="SA4",col_name="SA4")
        >>> australia = new GeoHierarchy(country,"Australia")
        >>> australia.add_region(region=state, parent_region=country)
        >>> australia.add_region(region=sa4, parent_region=state)
        >>> australia.add_type(sa4,"ASGS","Australian Statistical Geography Standard")
        >>> australia.type
        {'ASGS': 'Australian Statistical Geography Standard'}
        """

        # validate the arguments
        if type_id.strip() == "":
            raise ValueError("Type id cannot be empty")

        if not isinstance(region, Region):
            raise ValueError("`region` must be a Region")

        if type_id in self._type:
            raise ValueError("Duplicate type (" + type_id + ") found")

        if type_name.strip() != "":
            self._type[type_id] = type_name
        else:
            self._type[type_id] = type_id

        self._type_root[type_id] = region

    def add_region(self, region, parent_region):
        """
        Add a region as a child/sub region of another region
        Parameters
        ----------
        region:Region
            The sub region to be added as a child of the parent region
        parent_region:Region
            The direct upper-level of the region
        Examples
        --------
        >>> country = Region("Country")
        >>> state = Region("State",col_name="STATE")
        >>> australia = new GeoHierarchy(country,"Australia")
        >>> australia.add_region(region=state, parent_region=country)
        """

        # validate the arguments
        if not isinstance(region, Region):
            raise ValueError("`region` must be a Region")

        if not isinstance(parent_region, Region):
            raise ValueError("`parent_region` must be a Region")

        # if the parent region is the root of the hierarchy (country)
        # then add the new region as a sub region of the root
        if parent_region == self._country.region:
            self._country.add_child(self._Node(region))
        else:
            # find the node related to the parent_region
            parent_node = self._country.get_node_by_region(parent_region)
            if parent_node is not None:
                parent_node.add_child(self._Node(region))
            else:
                raise ValueError(str(parent_region) + " is not found")

    def get_regions_by_name(self, operator=None, name="", names=[], attribute=""):
        """
        Get all the relevant regions from the hierarchy based on the given parameters
        Parameters
        ----------
        operator: Operator
            use the operator to find all the upper/lower level regions
            from a particular region name.
            For instance: Country > State (Country gt State).
                          Use the 'gt' operator to search for the upper level of State
        name:string
            fill the name or short name of the regions in relations to operator parameter above
        names:list
            the list of region's name to look for
        attribute:string
            the region's attribute name that will be saved into the list
            (name, short_name, or col_name)
            if it's empty, then the list will store the object of the region
        Returns
        -------
        list
            list of regions or region's attribute
        Examples
        --------
        >>> country = Region("Country")
        >>> state = Region("State",col_name="STATE")
        >>> sa4 = Region("Statistica Area 4",short_name="SA4",col_name="SA4")
        >>> australia = new GeoHierarchy(country,"Australia")
        >>> australia.add_region(region=state, parent_region=country)
        >>> australia.add_region(region=sa4, parent_region=state)
        >>> regions = australia.get_regions_by_name()
        >>> for region in regions:
        >>>     print(region.name)
        Country
        State
        Statistica Area 4
        >>> regions = australia.get_regions_by_name(operator=le,name='State')
        >>> for region in regions:
        >>>     print(region.name)
        State
        Statistica Area 4
        >>> col_names = australia.get_regions_by_name(names=['State','SA4'], attribute='col_name')
        >>> for col_name in col_names:
        >>>     print(col_name)
        STATE
        SA4
        """

        # validate the arguments
        if operator is not None and len(names) > 0:
            raise ValueError(
                "Operator has to be empty if list of `names` is provided. vice versa."
            )

        if operator is not None and name.strip() == "":
            raise ValueError(
                "`name` must be provided if operator is not empty. vice versa."
            )

        if operator is not None:
            if operator not in [le, ge]:
                raise ValueError("Invalid operator value. Select one of 'le' or 'ge'")

        if attribute.strip() != "":
            if attribute not in ["name", "short_name", "col_name"]:
                raise ValueError(
                    "Invalid attribute value. Select one of name, short_name, or col_name"
                )

        regions = []
        # get multiple region. names are provided
        if len(names) > 0:
            for reg_name in names:
                node = self._country.get_node_by_name(reg_name)
                if node is not None:
                    if attribute.strip() == "":
                        regions.append(node.region)
                    else:
                        regions.append(getattr(node.region, attribute))
                else:
                    ValueError("Region is not found")
        # get all the corresponding regions based on the operator 'ge' or 'le'
        elif name != "":
            reference_node = self._country.get_node_by_name(name)
            if reference_node is not None:
                if operator in [ge]:
                    nodes = self._country.get_all_nodes(
                        lowest_region=reference_node.region
                    )
                    if attribute.strip() == "":
                        for node in nodes:
                            regions.append(node.region)
                    else:
                        for node in nodes:
                            regions.append(getattr(node.region, attribute))
                else:
                    nodes = reference_node.get_all_nodes()
                    if attribute.strip() == "":
                        for node in nodes:
                            regions.append(node.region)
                    else:
                        for node in nodes:
                            regions.append(getattr(node.region, attribute))
            else:
                ValueError("Region is not found")
        # get all the nodes
        else:
            nodes = self._country.get_all_nodes()
            if attribute.strip() == "":
                for node in nodes:
                    regions.append(node.region)
            else:
                for node in nodes:
                    regions.append(getattr(node.region, attribute))

        return regions

    def get_smallest_region_boundaries(self):
        """
        Get the smallest regional unit
        Returns
        -------
        Region
            the smallest regional unit
        Examples
        --------
        >>> country = Region("Country")
        >>> state = Region("State",col_name="STATE")
        >>> sa4 = Region("Statistica Area 4",short_name="SA4",col_name="SA4")
        >>> australia = new GeoHierarchy(country,"Australia")
        >>> australia.add_region(region=state, parent_region=country)
        >>> australia.add_region(region=sa4, parent_region=state)
        >>> australia.get_smallest_region_boundaries.name
        Statistica Area 4
        """
        return self._country.get_leaf_node().region

    class _Node:
        """
        The internal class for the hierarchy to encapsulate the Region
        instances into a Node for building the tree purposes.
        One Node contains one Region only.
        """

        __slot__ = ("_region", "_child")

        def __init__(self, region, child=[]):
            self._region = region

            if len(child) > 0:
                self._child = child
            else:
                self._child = []

        @property
        def region(self):
            """
            Return the region within the node
            """
            return self._region

        @property
        def child(self):
            """
            Return the list of the child node
            (similar to sub-region)
            """
            return self._child

        def add_child(self, child):
            """
            Add a child to the node
            (add sub-region to a region)
            """
            self._child.append(child)

        def get_node_by_region(self, region):
            """
            Searching for a node within its tree (the node itself and
            its subnode), based on the region's object.
            """
            if self._region == region:
                return self
            else:
                for child in self._child:
                    match = child.get_node_by_region(region)
                    if match is not None:
                        return match

            # print("Node "+str(region)+" is not found")
            return None

        def get_node_by_name(self, name):
            """
            Searching for a node within its tree (the node itself and
            its subnode), based on the region's name or short name.
            """
            if (
                self._region.name.lower() == name.lower()
                or self._region.short_name.lower() == name.lower()
            ):
                return self
            else:
                for child in self._child:
                    match = child.get_node_by_name(name)
                    if match is not None:
                        return match

            return None

        def get_all_nodes(self, lowest_region=None):

            nodes = []
            if len(self._child) == 0:
                return [self]
            elif self._region == lowest_region:
                return [self]
            else:
                nodes += [self]
                for child in self._child:
                    if child not in nodes:
                        nodes += child.get_all_nodes(lowest_region)

            return nodes

        def get_leaf_node(self):
            if len(self._child) == 0:
                return self
            else:
                for child in self._child:
                    match = child.get_leaf_node()
                    if match is not None:
                        return match

        def __str__(self):
            return str(self._region + ". child:" + str(len(self._child)))

        def __eq__(self, other):
            return self._region == other._region



"""
Data structure for the regional unit
"""


class Region:
    """
    The Region class represents a unit of area in the country.
    Each region has a unique name and a corresponding column 
    in the reference dataset.
    """

    __slots__ = ("_name", "_short_name", "_col_name")

    def __init__(self, name, short_name="", col_name=""):
        # region name cannot be empty
        if name is None:
            raise TabError("`name` must not be None")

        self._name = name
        self._short_name = short_name
        self._col_name = col_name

    @property
    def name(self):
        """
        Return the name of the administrative area or statistical area level
        Returns
        -------
        string
            The name of the area unit
        Examples
        --------
        The area name can be set initially when calling the constructor.
        >>> state = new Region('State',col_name='STATE')
        >>> state.name
        'State'
        """
        return self._name

    @name.setter
    def name(self, value):
        """
        Set or rename the name of the administrative area or statistical area level
        Parameters
        ----------
        value: string
            The name of the area unit
        Examples
        --------
        >>> state.name = 'State'
        >>> state.name
        'State'
        """
        self._name = value

    @property
    def short_name(self):
        """
        Return the short name or abbreviation of the administrative area or
        statistical area level
        Returns
        -------
        string
            The short name of the area unit
        Examples
        --------
        The area's short name can be set initially when calling the constructor.
        >>> sa1 = new Region('Statistical Area 1',short_name='SA1',col_name='SA1')
        >>> sa1.short_name
        'SA1'
        """
        return self._short_name

    @short_name.setter
    def short_name(self, value):
        """
        Set or rename the short name or abbreviation of the administrative area or
        statistical area level
        Parameters
        ----------
        value: string
            The short name of the area unit
        Examples
        --------
        >>> sa1.short_name = 'SA-1'
        >>> sa1.short_name
        'SA-1'
        """
        self._id = value

    @property
    def col_name(self):
        """
        Return the corresponding column name of the administrative area or
        statistical area level which is stored in the reference dataset
        Returns
        -------
        string
            The column name of the area unit
        Examples
        --------
        The area's column name can be set initially when calling the constructor.
        >>> sa2 = new Region('Statistical Area 2',short_name='SA2',col_name='SA2')
        >>> sa2.col_name
        'SA2'
        """
        return self._col_name

    @col_name.setter
    def col_name(self, value):
        """
        Set or rename the column name of the administrative area or
        statistical area level
        Parameters
        ----------
        value: string
            The short name of the area unit
        Examples
        --------
        >>> sa2.col_name = 'SA2_2016'
        >>> sa2.col_name
        'SA2_2016'
        """
        self._col_name = value

    def __str__(self):
        return self._name

    def __eq__(self, other):
        return (
            isinstance(other, Region)
            and self._short_name == other._short_name
            and self._name == other._name
        )

    def __ne__(self, other):
        return not (self == other)



from urllib import request
import re
import json
import os
import sys
from colorama import Fore, Style
import argparse
import signal

CWD = os.path.abspath(os.getcwd())

color_code = {
    "default": "",
    "red": Fore.RED,
    "green": Style.BRIGHT + Fore.GREEN,
}
# ANSI terminal control sequence to erase the current line
erase_line = "\x1b[2K"


def print_text(
    text, color="default", in_place=False, **kwargs
):  # type: (str, str, bool, any) -> None
    """
    Print text to console
    Parameters
    ----------
    text : string
           text to print
    color : string
            it can be one of "red" or "green", or "default"
    in_place : boolean
               whether to erase previous line and print in place
    **kwargs : dict, optional
             : other keywords passed to built-in print
    Returns
    -------
    None
    """
    if in_place:
        print("\r" + erase_line, end="")
    print(color_code[color] + text + Style.RESET_ALL, **kwargs)


def create_url(url):
    """
    Produce a URL that is compatible with Github's REST API from the input url
    This can handle blob or tree paths.
    Parameters
    ----------
    url : string
          url to the data directory in Github repository
    Returns
    -------
    apiurl      : string
                  Github API url
    downloadurl : string
                  Download directory

    """
    repo_only_url = re.compile(
        r"https:\/\/github\.com\/[a-z\d](?:[a-z\d]|-(?=[a-z\d])){0,38}\/[a-zA-Z0-9]+$"
    )
    re_branch = re.compile("/(tree|blob)/(.+?)/")

    # Check if the given url is a url to a GitHub repo.
    # If it is, inform the user to use 'git clone' to download it
    if re.match(repo_only_url, url):

        print_text(
            "✘ The given url is a complete repository. "
            "Use 'git clone' to download the repository",
            "red",
            in_place=True,
        )
        sys.exit()

    # Extract the branch name from the given url (e.g master)
    branch = re_branch.search(url)
    download_dirs = url[branch.end() :]
    api_url = (
        url[: branch.start()].replace("github.com", "api.github.com/repos", 1)
        + "/contents/"
        + download_dirs
        + "?ref="
        + branch.group(2)
    )
    return api_url, download_dirs


def download_data(country="Australia", output_dir=CWD):
    """
    download the files and directories and sub-directories in repo_url.

    Parameters
    ----------
    country : str
              country name which will be sub-directory name example - data/Australia/
    Return
    ------
    total_file  : int
                  number of total files downloaded
    """

    # This is the temporary place to host of data files.
    # Example - data/Australia/*.parquet
    repo_url = (
        "https://github.com/uts-mdsi-ilab2-synergy/addrmatcher/tree/main/data/"
        + country
        + "/"
    )

    # Generate the url which returns the JSON data
    api_url, download_dirs = create_url(repo_url)

    # Handle the directory path, Sync the path in Git and the path in local
    if len(download_dirs.split(".")) == 0:
        dir_out = os.path.join(output_dir, download_dirs)
    else:
        dir_out = os.path.join(output_dir, "/".join(download_dirs.split("/")[:-1]))

    # Make a directory in the local with the name which is taken from the actual repo
    os.makedirs(dir_out, exist_ok=True)

    # Open the URL
    try:
        opener = request.build_opener()
        opener.addheaders = [("User-agent", "Mozilla/5.0")]
        request.install_opener(opener)
        response = request.urlretrieve(api_url)
    except KeyboardInterrupt:
        # when CTRL+C is pressed during the execution of this script,
        # bring the cursor to the beginning,
        # erase the current line, and dont make a new line
        print_text("✘ Got interrupted", "red", in_place=True)
        sys.exit()

    # total files count
    total_files = 0

    # Writing into file
    with open(response[0], "r") as f:
        data = json.load(f)
        # getting the total number of files so that we
        # can use it for the output information later
        total_files += len(data)

        # If the data is a single file, download it as one.
        if isinstance(data, dict) and data["type"] == "file":
            try:
                # download the file
                opener = request.build_opener()
                opener.addheaders = [("User-agent", "Mozilla/5.0")]
                request.install_opener(opener)
                request.urlretrieve(
                    data["download_url"], os.path.join(dir_out, data["name"])
                )
                # bring the cursor to the beginning,
                # erase the current line, and dont make a new line
                print_text(
                    "Downloaded: " + Fore.WHITE + "{}".format(data["name"]),
                    "green",
                    in_place=True,
                )
                return total_files

            except KeyboardInterrupt:
                # when CTRL+C is pressed during the execution of this script,
                # bring the cursor to the beginning,
                # erase the current line, and dont make a new line
                print_text("✘ Got interrupted", "red", in_place=False)
                sys.exit()

        # If Data is a directory which contains a list of file, not a single file
        for file in data:
            file_url = file["download_url"]
            file_name = file["name"]

            path = file["path"]
            dirname = os.path.dirname(path)

            if dirname != "":
                os.makedirs(os.path.dirname(path), exist_ok=True)
            else:
                pass

            if file_url is not None:
                try:
                    opener = request.build_opener()
                    opener.addheaders = [("User-agent", "Mozilla/5.0")]
                    request.install_opener(opener)
                    # download the file
                    request.urlretrieve(file_url, path)

                    # bring the cursor to the beginning,
                    # erase the current line, and dont make a new line
                    print_text(
                        "Downloaded" + Fore.WHITE + "{}".format(file_name),
                        "green",
                        in_place=False,
                        end="\n",
                        flush=True,
                    )

                except KeyboardInterrupt:
                    # when CTRL+C is pressed during
                    # the execution of this script,
                    # bring the cursor to the beginning,
                    # erase the current line, and dont make a new line
                    print_text("✘ Got interrupted", "red", in_place=False)
                    sys.exit()
            else:
                download_data(file["html_url"], dir_out)

    return total_files


def download():
    """
    Trigger the download_data function and read the argument from user's CLI
    """

    if sys.platform != "win32":
        # disbale CTRL+Z
        signal.signal(signal.SIGTSTP, signal.SIG_IGN)

    # Parse the directory i.e country name
    parser = argparse.ArgumentParser(
        description="Download directories/folders from GitHub"
    )

    # parser.add_argument(
    #     "country",
    #     type=str,
    #     default="Australia",
    #     help="Name of the country for which the reference data to download",
    # )

    # args = parser.parse_args()
    # country = args.country

    download_data()
    print_text("✔ Download complete", "green", in_place=True)


if __name__ == "__main__":
    download()



from rapidfuzz import fuzz
from rapidfuzz.string_metric import jaro_similarity, jaro_winkler_similarity
import pandas as pd
import re
import os
import glob
from pyarrow import fs
import pyarrow.parquet as pq
from sklearn.neighbors import BallTree
import numpy as np


class GeoMatcher:

    __slot__ = (
        "_hierarchy",
        "_file_location",
        "_index_data",
        "_filename",
        "_street_code_dict",
    )

    def __init__(self, hierarchy, file_location=""):
        self._hierarchy = hierarchy

        # if no file location provided, look for the dataset in the default folder: data/[country]
        if file_location.strip() == "":
            if os.path.isdir(os.path.join("data", self._hierarchy.name)):
                self._file_location = os.path.join("data", self._hierarchy.name)
            else:
                raise ValueError(
                    "Folder names that contain the index file can't be found: "
                    + os.path.join("data", self._hierarchy.name)
                )
        else:
            if os.path.isdir(file_location):
                self._file_location = file_location
            else:
                raise ValueError(
                    "Folder names that contain the index file can't be found: "
                    + file_location
                )

        # get all the parquet filenames within the folder
        self._filename = glob.glob(os.path.join(self._file_location, "*.parquet"))

        # init
        index_file = "index.parquet"

        # check if the index file exists
        if os.path.join(self._file_location, index_file) not in self._filename:
            raise ValueError(
                "Index file ("
                + index_file
                + ") can't be found in: "
                + self._file_location
            )

        # read the index file
        self._index_data = pd.read_parquet(
            os.path.join(self._file_location, index_file)
        )

        # check the availability of required column name
        idx_columns = ["IDX", "ADDRESS", "FILE_NAME"]
        if not set(idx_columns).issubset(self._index_data.columns):
            raise ValueError(
                "The required columns can't be found in the index file: "
                + str(set(idx_columns) - set(self._index_data.columns))
            )

        # remove index file from the lists
        self._filename.remove(os.path.join(self._file_location, index_file))

        # check parquet file schema (ensure all of the required columns are exist)
        # get the regions that users selected
        all_regions = self._hierarchy.get_regions_by_name(attribute="col_name")
        all_columns = list(filter(None, all_regions))

        for file in self._filename:
            pq_columns = pq.read_schema(file).names
            if not set(all_columns).issubset(pq_columns):
                raise ValueError(
                    "The required columns "
                    + str(set(all_columns) - set(pq_columns))
                    + " can't be found in the parquet file: "
                    + file
                )

        # define the dictionary for street code normalization
        self._street_code_dict = {
            "ALLY": "ALLEY",
            "ALY": "ALLEY",
            "ARC": "ARCADE",
            "AVE": "AVENUE",
            "AV": "AVENUE",
            "BLTWY": "BELTWAY",
            "BVD": "BOULEVARD",
            "BYPA": "BYPASS",
            "CCT": "CIRCUIT",
            "CL": "CLOSE",
            "CRN": "CORNER",
            "CT": "COURT",
            "CRES": "CRESCENT",
            "CSWY": "CAUSEWAY",
            "CDS": "CUL-DE-SAC",
            "DR": "DRIVE",
            "ESP": "ESPLANADE",
            "GRN": "GREEN",
            "GR": "GROVE",
            "HWY": "HIGHWAY",
            "JNC": "JUNCTION",
            "LN": "LANE",
            "LANE": "LANE",
            "LINK": "LINK",
            "MEWS": "MEWS",
            "PDE": "PARADE",
            "PKWY": "PARKWAY",
            "PL": "PLACE",
            "RDGE": "RIDGE",
            "RD": "ROAD",
            "SQ": "SQUARE",
            "ST": "STREET",
            "TCE": "TERRACE",
            "TPKE": "TURNPIKE",
            "WAY": "WAY",
        }

    def _remove_street_number(self, address):
        """
        Remove the street number, lot/unit/level number, or similar attributes
        from the address
        Parameter
        ---------
        address: string
            The physical address' text
        Return
        ------
            An address without the street number
        """

        # initiate the result
        no_number_address = address
        while True:
            match = re.search(r"[0-9]+[a-zA-Z,]*\s", address)
            if not match:
                break

            pos = match.span()[1]
            address = address[pos:]
            no_number_address = address.upper()

        # change to upper case
        no_number_address = no_number_address.replace(",", "").strip().upper()
        # replace the street suffix abbreviation with the street suffix name
        # and remove the extra spaces
        address_parts = no_number_address.split()

        if address_parts[0] == "ST":
            address_parts.pop(0)
            return "ST " + " ".join(
                [self._street_code_dict.get(item, item) for item in address_parts]
            )
        else:
            return " ".join(
                [self._street_code_dict.get(item, item) for item in address_parts]
            )

    def _cleaning_match_with_index(self, no_number_address):
        """
        Return similar addresses of the no_number_address with
        the index file based on postcode and state or
        locality/suburb or street name.
        Parameter
        ---------
        no_number_address: string
            The physical address without the street number,
            lot/unit/level number, or similar attribute
        Returns
        -------
            A matched address
        """

        # initiate the dataframe
        matched_df = pd.DataFrame()

        # search State and Postcode and extract - example QLD 410
        match = re.search(
            r"\s((?:NSW|VIC|QLD|TAS|WA|SA|NT|ACT))\s([0-9]{4})$", no_number_address
        )
        if not match:
            return matched_df

        state, postcode = match.group(1), match.group(2)

        # firstly, Filter for the rows from the Index File based on same Postcode and State
        matched_postcode_state = self._index_data[
            (self._index_data["POSTCODE"] == postcode)
            & (self._index_data["STATE"] == state)
        ]
        if matched_postcode_state.empty:
            return matched_df

        # secondly, Filter further for same Suburbs,
        # get the Distinct Suburb Name List first - "LOCALITY_NAME"

        suburb_list = matched_postcode_state["LOCALITY_NAME"].unique()
        suburbs = [sub for sub in suburb_list if sub in no_number_address]
        if not suburbs:
            return matched_df

        # get the boolean array - of rows matching (LOCALITY_NAME)
        # with filtered suburbs in the above list
        index = matched_postcode_state["LOCALITY_NAME"].str.contains("|".join(suburbs))
        if not any(index):
            return matched_df

        # find Street and Street code from no_number_address to further matching by street name
        # sample match output - <re.Match object; span=(29, 46), match='WEST END QLD 4101'>
        match = re.search(
            f"(?:{'|'.join(suburbs)})\s+{state}\s+{postcode}", no_number_address
        )
        if not match:
            return matched_df
        # extract (unit, street) from the whole address string
        # - UNIT 415 21 ABC STREET  from "UNIT 415 21 ABC STREET WEST END QLD 4101"
        street_string = no_number_address[: match.span()[0]]

        # thirdly, Filter further by Street Name within the filtered Suburb
        subb_df = matched_postcode_state[index]
        unique_street_name_list = subb_df["STREET_NAME"].unique()
        street_name = [
            street_name
            for street_name in unique_street_name_list
            if street_name in street_string
        ]

        if (
            street_name
        ):  # if street name does not exist in the data frame, suburbs maybe incorrect
            matched_df = subb_df[subb_df["STREET_NAME"].isin(street_name)]

        else:
            # find within the same post code and state again for street name
            street_name = [
                street_name
                for street_name in unique_street_name_list
                if street_name in street_string
            ]
            matched_df = matched_postcode_state[
                matched_postcode_state["STREET_NAME"].isin(street_name)
            ]

        return matched_df

    def _cleaning_address(self, no_number_address):
        """
        Return the clean address.
        The function will revise the locality/suburb or other attributes
        if necessary.
        Parameter
        ---------
        no_number_address: string
            The physical address without the street number,
            lot/unit/level number, or similar attribute
        Return
        ------
            A clean address
        """

        matched_df = self._cleaning_match_with_index(no_number_address)

        if matched_df.empty:
            return no_number_address
        else:

            match = re.search(
                "\s((?:NSW|VIC|QLD|TAS|WA|SA|NT|ACT))\s([0-9]{4})$", no_number_address
            )
            if match:
                state, postcode = match.group(1), match.group(2)

            match = re.search(
                f"(?:({'|'.join(matched_df['STREET_NAME'].unique())}))\s(.*)\s((:?{'|'.join(matched_df['LOCALITY_NAME'].unique())}))",
                no_number_address,
            )

            if not match:
                match = re.search(
                    f"(?:({'|'.join(matched_df['STREET_NAME'].unique())}))\s(\S*)\s",
                    no_number_address,
                )
                if not match:
                    return no_number_address

                street_name = match.group(1)
                street_code = match.group(2)

                match = re.search(f"{street_code}\s(.+)\s{state}", no_number_address)
                # Suburb maybe incorrectly entered
                suburb = match.group(1) if match else None
            else:
                street_name, street_code, suburb = (
                    match.group(i + 1) for i in range(3)
                )

            if suburb:
                # Get the index of filter based on Street and Suburb name
                index = (matched_df["STREET_NAME"] == street_name) & (
                    matched_df["LOCALITY_NAME"] == suburb
                )
                # If no exact match found, try to find matching string with Street name and Suburb
                index = (
                    index
                    if any(index)
                    else (matched_df["STREET_NAME"].str.contains(street_name))
                    & (matched_df["LOCALITY_NAME"].str.contains(suburb))
                )

                # if there is no matching suburbs found, use street name only to filter
                matched_df = (
                    matched_df[index]
                    if any(index)
                    else matched_df[
                        (matched_df["STREET_NAME"].str.contains(street_name))
                    ]
                )

            else:
                matched_df = matched_df[(matched_df["STREET_NAME"] == street_name)]

            if matched_df.empty:
                return no_number_address
            if matched_df.shape[0] == 1:
                return matched_df["ADDRESS"].iloc[0]

            # Covert street code to long form postcode
            street_code = (
                self._street_code_dict[street_code]
                if street_code in self._street_code_dict
                else street_code
            )
            matched_df = matched_df[matched_df["STREET_TYPE_CODE"] == street_code]

            if matched_df.empty:
                return no_number_address
            if matched_df.shape[0] == 1:
                return matched_df["ADDRESS"].iloc[0]
            return no_number_address

    def get_region_by_address(
        self,
        address,
        similarity_threshold=0.9,
        nlargest=1,
        regions=[],
        operator=None,
        region="",
        address_cleaning=False,
        string_metric="levenshtein",
    ):
        """
        perform address based matching and return the corresponding region
        e.g. administrative level or statistical area

        Parameters
        ----------
        address:string
            The complete physical address
        similarity_threshold:float
            The minimum similarity ratio ranges between 0 and 1 (default = 0.9)
        nlargest:int
            The number of the addresses to be returned by the function.
            If nlargest = 1, then the function will return the top similarity only
            (default = 1)
        regions:list
            Specify the name of the regions to be returned by the function
        operator:list
            use the operator to find all the upper/lower level regions
            from a particular region name.
            For instance: Country >= State (Country ge State).
                          Use the 'ge' operator to search for the upper level of State
                          (and itself)
        region:string
            fill the name or short name of the regions in relations to operator parameter above
        address_cleaning:boolean
            perform data cleaning on the address, for instance: revise invalid suburb name
            (currently, only applied to Australian addresses)
        string_metric:string
            The name of the edit distance algorithm used.
            Select one of 'levenshtein','jaro', or 'jaro_winkler'
        Returns
        -------
            the dictionary of the matched adddresses
        Examples
        --------
        >>> matcher = GeoMatcher(AUS)
        >>> matched = matcher.get_region_by_address("25 Darnley Street,
                      Braybrookt, VIC 3019", similarity_threshold = 0.95)
        >>> matched
        {'MB_CODE_2016': ['20375120000'],
         'SA4_NAME_2016': ['Melbourne - West'],
         'SA3_NAME_2016': ['Maribyrnong'],
         'SA2_NAME_2016': ['Braybrook']
         'SA1_7DIGITCODE_2016': ['2134703'],
         'STATE': ['VIC'],
         'RATIO': [0.9841269841269842],
         'SSC_NAME_2016': ['Braybrook'],
         'LGA_NAME_2016': ['Maribyrnong (C)'],
         'FULL_ADDRESS': ['25 DARNLEY STREET BRAYBROOK VIC 3019']}
        """

        if string_metric not in ["levenshtein", "jaro", "jaro_winkler"]:
            raise ValueError(
                "String metric is unknown. Select one of 'levenshtein','jaro','jaro_winkler'"
            )

        if similarity_threshold < 0:
            raise ValueError("Similarity threshold has to be larger than 0")

        if nlargest < 1:
            raise ValueError("The number of returned records must be at least 1")

        # initiate the result
        addresses = pd.DataFrame()

        clean_address = self._remove_street_number(address)

        if address_cleaning:
            # perform further cleaning
            clean_address = self._cleaning_address(clean_address)

        # print(clean_address)

        # match with the index
        if (self._index_data is not None) and (self._index_data.shape[0] > 0):
            # no clean address found
            clean_address_idx = self._index_data[
                self._index_data["ADDRESS"] == clean_address
            ]

            if clean_address_idx.shape[0] == 0:
                # calculate the distance (Levenshtein Distance) between the
                # input address (without street number) and the index
                # [all special characters are removed]
                if string_metric == "levenshtein":
                    self._index_data["RATIO"] = self._index_data["ADDRESS"].apply(
                        lambda x: fuzz.ratio(
                            re.sub(r"[\W_]+", "", clean_address),
                            re.sub(r"[\W_]+", "", x.upper()),
                        )
                        / 100.0
                    )
                elif string_metric == "jaro":
                    self._index_data["RATIO"] = self._index_data["ADDRESS"].apply(
                        lambda x: jaro_similarity(
                            re.sub(r"[\W_]+", "", clean_address),
                            re.sub(r"[\W_]+", "", x.upper()),
                        )
                        / 100.0
                    )
                else:
                    self._index_data["RATIO"] = self._index_data["ADDRESS"].apply(
                        lambda x: jaro_winkler_similarity(
                            re.sub(r"[\W_]+", "", clean_address),
                            re.sub(r"[\W_]+", "", x.upper()),
                        )
                        / 100.0
                    )

                # get the index with the largest similarity
                largest_idx = self._index_data.nlargest(1, "RATIO")

                if largest_idx["RATIO"].values[0] < similarity_threshold:
                    return {}

                # first, check the filename it it's available
                parquet_filename = largest_idx["FILE_NAME"].values[0]
                parquet_idx = largest_idx["IDX"].values[0]
            else:

                parquet_filename = clean_address_idx["FILE_NAME"].values[0]
                parquet_idx = clean_address_idx["IDX"].values[0]

            if os.path.isfile(os.path.join(self._file_location, parquet_filename)):
                # read the parquet file where the IDX and address are stored
                address_parquet = pq.read_table(
                    os.path.join(self._file_location, parquet_filename),
                    filesystem=fs.LocalFileSystem(),
                    filters=[("IDX", "=", parquet_idx)],
                ).to_pandas()

                # calculate the distance (Levenshtein Distance) between the
                # input address (with street number) and the entire addresses
                # reference dataset
                # [all special characters are removed]
                if string_metric == "levenshtein":
                    address_parquet["RATIO"] = address_parquet["FULL_ADDRESS"].apply(
                        lambda x: fuzz.ratio(
                            re.sub(r"[\W_]+", "", address.upper()),
                            re.sub(r"[\W_]+", "", x.upper()),
                        )
                        / 100.0
                    )
                elif string_metric == "jaro":
                    address_parquet["RATIO"] = address_parquet["FULL_ADDRESS"].apply(
                        lambda x: jaro_similarity(
                            re.sub(r"[\W_]+", "", address.upper()),
                            re.sub(r"[\W_]+", "", x.upper()),
                        )
                        / 100.0
                    )
                else:
                    address_parquet["RATIO"] = address_parquet["FULL_ADDRESS"].apply(
                        lambda x: jaro_winkler_similarity(
                            re.sub(r"[\W_]+", "", address.upper()),
                            re.sub(r"[\W_]+", "", x.upper()),
                        )
                        / 100.0
                    )

                # if similarity score is larger then the threshold,
                # there is a possibility the addresses are similar
                # will need to select the highest score later on
                addresses = address_parquet[
                    address_parquet["RATIO"] >= similarity_threshold
                ]

                # get the regions that users selected
                selected_regions = self._hierarchy.get_regions_by_name(
                    operator=operator, name=region, names=regions
                )

                # get the columns only
                selected_columns = []
                for reg in selected_regions:
                    if reg not in selected_columns:
                        selected_columns.append(reg.col_name)

                # deleted later
                selected_columns.append("FULL_ADDRESS")
                selected_columns.append("RATIO")

                # remove empty element, if exists
                selected_columns = list(filter(None, selected_columns))
                selected_columns = list(set(selected_columns))

                # if there are possible similar address found
                if addresses.shape[0] > 0:
                    # return the most similar address only
                    if nlargest == 1:

                        # sort the addresses based on the similarity score
                        addresses = addresses.sort_values(
                            by="RATIO", ascending=False
                        ).reset_index(drop=True)

                        return addresses.head(1)[selected_columns].to_dict(
                            orient="list"
                        )

                    # return all the similar addresses
                    else:

                        return (
                            addresses[selected_columns]
                            .nlargest(nlargest, "RATIO")
                            .sort_values(by="RATIO", ascending=False)
                            .to_dict(orient="list")
                        )
                else:
                    return {}

            else:
                raise ValueError("The address file can't be found: " + parquet_filename)

        else:
            raise ValueError(
                "No index records found. Make sure the initiation process is succeeded"
            )

    def _load_parquet(self, lat, lon, distance):
        """
        load the rows in the parquet file meeting the condition
        the condition is to ensure the LATITUDE and LONGITUDE are within a distance from the
        argument lat and lon

        Parameters
        ----------
        lat:float
            latitude
        lon:float
            longitude
        distance:integer
                define what the minimum km of distance from the argument lat and lon
        Returns
        -------
        a panda dataframe
        """

        local = fs.LocalFileSystem()
        df = pq.read_table(
            self._filename,
            filesystem=local,
            filters=[
                ("LATITUDE", ">=", lat - distance),
                ("LATITUDE", "<=", lat + distance),
                ("LONGITUDE", ">=", lon - distance),
                ("LONGITUDE", "<=", lon + distance),
            ],
        ).to_pandas()

        return df

    def get_region_by_coordinates(
        self, lat, lon, n=1, km=1, regions=[], operator=None, region=""
    ):
        """
        perform coordinate_based matching and return the corresponding regions in a dictionary
        e.g. administrative level or statistical area

        Parameters
        ----------
        lat:float
            latitude
        lon:float
            longitude
        n:integer
            the number of nearest addresses to be returned by the function.
        km:integer
            the nearest addresses will be searched from the input coordinates
            point within the argument kilometer radius
        Returns
        -------
            a dictionary of addresses with statistical and administrative regions
        Examples
        --------
        >>> matcher = GeoMatcher(AUS)
        >>> matched = matcher.get_region_by_address(-26.657299,153.094955)
        >>> matched
        {'FULL_ADDRESS': ['8 32 SECOND AVENUE MAROOCHYDORE QLD 4558'],
         'LATITUDE': [-26.6572865955204],
         'LONGITUDE': [153.09496396875],
         'LGA_NAME_2016': ['Sunshine Coast (R)'],
         'SSC_NAME_2016': ['Maroochydore'],
         'SA4_NAME_2016': ['Sunshine Coast'],
         'SA3_NAME_2016': ['Maroochy'],
         'SA2_NAME_2016': ['Maroochydore - Kuluin'],
         'SA1_7DIGITCODE_2016': ['3142707'],
         'MB_CODE_2016': ['30563074700'],
         'DISTANCE': [0.0016422183328786543]}
        """

        min_distance = 0
        # 1 lat equals 110.574km
        distance = (km if km else 1) / 110.574

        # 1. Ensure lat/lon within the country's geo boundary range
        if len(self._hierarchy.coordinate_boundary) != 4:
            raise ValueError("The country's geo boundary is not available")

        within_range = (
            self._hierarchy.coordinate_boundary[0]
            <= lat
            <= self._hierarchy.coordinate_boundary[1]
        ) and (
            self._hierarchy.coordinate_boundary[2]
            <= lon
            <= self._hierarchy.coordinate_boundary[3]
        )
        if not within_range:
            raise ValueError(
                "The latitude input should be within "
                + self._hierarchy.coordinate_boundary[0]
                + "and"
                + self._hierarchy.coordinate_boundary[1]
                + " and longitude input must be within  "
                + self._hierarchy.coordinate_boundary[2]
                + "and"
                + self._hierarchy.coordinate_boundary[3]
            )

        # 2. Make the first load of GNAF dataset
        gnaf_df = self._load_parquet(lat, lon, distance)

        # 2.a If the desired count of addresses not exist, increase the radius
        while gnaf_df.shape[0] < n:
            min_distance = distance
            distance *= 2

            gnaf_df = self._load_parquet(lat, lon, distance)

        # 2.b Keep reducing the size of rows if more than 10k adddresses
        # are found within the radius
        # Take the median distance to reduce
        # This is to limit the number of datapoint to build the Ball tree in the next step
        while gnaf_df.shape[0] >= n + 10000:
            middle_distance = (distance - min_distance) / 2

            temp_df = gnaf_df[
                gnaf_df.LATITUDE.between(lat - middle_distance, lat + middle_distance)
                & gnaf_df.LONGITUDE.between(
                    lon - middle_distance, lon + middle_distance
                )
            ]
            # If no record are found, quit the iteration
            if temp_df.shape[0] < 1:
                break
            gnaf_df = temp_df
            distance = middle_distance

        # 3. Build the Ball Tree and Query for the nearest within k distance
        ball_tree = BallTree(
            np.deg2rad(gnaf_df[["LATITUDE", "LONGITUDE"]].values), metric="haversine"
        )

        distances, indices = ball_tree.query(
            np.deg2rad(np.c_[lat, lon]), k=min(n, gnaf_df.shape[0])
        )
        # Get indices of the search result, Extract pid and calculate distance(km)
        indices = indices[0].tolist()
        pids = gnaf_df.ADDRESS_DETAIL_PID.iloc[indices].tolist()
        distance_map = dict(zip(pids, [distance * 6371 for distance in distances[0]]))

        # 4. Filter the GNAF dataset by address_detail_pid
        bool_list = gnaf_df["ADDRESS_DETAIL_PID"].isin(pids)
        final_gnaf_df = gnaf_df[bool_list]

        final_gnaf_df.loc[:, "DISTANCE"] = final_gnaf_df["ADDRESS_DETAIL_PID"].map(
            distance_map
        )

        return final_gnaf_df.sort_values("DISTANCE").to_dict(orient="list")
